/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.fmap.dal.feature;

import org.gvsig.fmap.dal.DataTypes;
import org.gvsig.tools.evaluator.Evaluator;

/**
 * This interface represents a FeatureType in editable state. 
 * To edit a FeatureType you have to obtain its instance of 
 * EditableFeatureType and then perform editing operations on it.
 * 
 * Once you have completed the editing you can save the changes to the original 
 * FeatureType. This is the only way to edit a FeatureType.
 */
public interface EditableFeatureType extends FeatureType {

	/**
	 * Sets the default geometry attribute name
	 * 
	 * @param name
	 * 			string containing the default geometry attribute name
	 */
	public void setDefaultGeometryAttributeName(String name);

	/**
	 * Removes an attribute from this EditableFeatureType, given
	 * a reference to its descriptor.
	 * 
	 * @param attribute
	 * 				descriptor of the attribute to remove
	 * 
	 * @return
	 * 		true if the attribute was removed successfully, false if not.
	 */
	public boolean remove(EditableFeatureAttributeDescriptor attribute);

	/**
	 * Removes an attribute given its name
	 * 
	 * @param name
	 * 			string containing the name of the attribute to be removed
	 * @return
	 * 		
	 */
	public Object remove(String name);

	/**
	 * Removes an attribute given its index
	 * 
	 * @param index
	 * 			position of the attribute to be removed
	 * 
	 * @return
	 * 		
	 */
	public Object remove(int index);

	/**
	 * Adds an attribute to this EditableFeatureType. 
	 * @param name
	 * 			string containing the name of the attribute
	 * @param type
	 * 			data type of the attribute (one from {@link DataTypes})
	 * 
	 * @return a new EditableFeatureAttributeDescriptor
	 */
	public EditableFeatureAttributeDescriptor add(String name, int type);

	/**
	 * Adds an attribute to this EditableFeatureType.
	 *  
	 * @param name
	 * 			string containing the name of the attribute
	 * 
	 * @param type
	 * 			data type of the attribute (one from {@link DataTypes})
	 * 
	 * @param size
	 * 			size of the attribute.
	 * 
	 * @return a new EditableFeatureAttributeDescriptor
	 */
	public EditableFeatureAttributeDescriptor add(String name, int type,
			int size);

	/**
	 * Adds a calculated attribute to this EditableFeatureType.
	 *  
	 * @param name
	 * 			string containing the name of the attribute
	 * 
	 * @param type
	 * 			data type of the attribute (one from {@link DataTypes})
	 * 
	 * @param evaluator
	 * 			an evaluator containing the desired expression
	 * 
	 * @return a new EditableFeatureAttributeDescriptor
	 */	
	public EditableFeatureAttributeDescriptor add(String name, int type,
			Evaluator evaluator);

	public EditableFeatureAttributeDescriptor add(String name, int type,
			FeatureAttributeEmulator emulator);
	/**
	 * Returns the associated FeatureType.
	 *  
	 * @return the associated FeatureType
	 */
	public FeatureType getSource();

	/**
	 * Returns a copy of the associated FeatureType.
	 * 
	 * @return associated FeatureType
	 */
	public FeatureType getNotEditableCopy();

	/**
	 * Sets whether this EditableFeatureType has an OID.
	 * An OID is a unique serializable reference to a feature
	 * (a primary key of sorts that may or may not be defined 
	 * by the store). If the store does not define this OID then
	 * it will be generated by the feature reference itself.
	 * 
	 * Its main use is to provide a way to persist data associated 
	 * to a feature by this OID.
	 * 
	 * @param hasOID true if it has an OID, or false if not.
	 */
	public void setHasOID(boolean hasOID);
	
	/**
	 * Sets the default time attribute name
     * 
     * @param name
     *          string containing the default time attribute name
	 */
	public void setDefaultTimeAttributeName(String name);
        
	/**
	 * Returns a {@link FeatureAttributeDescriptor} given the attribute name,
	 * or null if an attribute with the given name does not exist.
	 *
	 * @param name
	 * 			of the attribute
	 *
	 * @return
	 * 		a {@link FeatureAttributeDescriptor}
	 */
	public EditableFeatureAttributeDescriptor getEditableAttributeDescriptor(String name);

	/**
	 * Returns a {@link FeatureAttributeDescriptor} given the attribute index.
	 *
	 * @param index
	 * 			of the attribute
	 *
	 * @return
	 * 		a {@link FeatureAttributeDescriptor}
	 */
	public EditableFeatureAttributeDescriptor getEditableAttributeDescriptor(int index);
}
