package org.gvsig.fmap.dal.store.csv.simplereaders;

import java.io.FileReader;
import java.io.IOException;
import java.util.List;
import org.apache.commons.lang3.StringUtils;
import org.gvsig.fmap.dal.store.csv.CSVStoreParameters;
import org.gvsig.tools.dynobject.DynObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.supercsv.comment.CommentStartsWith;
import org.supercsv.io.CsvListReader;
import org.supercsv.prefs.CsvPreference;
import org.supercsv.quote.QuoteMode;

public class CSVReader implements SimpleReader {

    //
    // http://supercsv.sourceforge.net/examples_reading.html
    // http://supercsv.sourceforge.net/apidocs/index.html
    //
    private static final Logger logger = LoggerFactory.getLogger(CSVReader.class);
    CsvListReader reader;
    private CSVStoreParameters parameters;

    public CSVReader(FileReader in, CSVStoreParameters parameters) {
        this.parameters = parameters;
        CsvPreference preferences = getCSVPreferences();
        this.reader = new CsvListReader(in, preferences);
    }    
    
    public CSVReader(CSVStoreParameters parameters) {
        this.parameters = parameters;
        CsvPreference preferences = getCSVPreferences();
        this.reader = null;
    }

    public CSVStoreParameters getParameters() {
        return this.parameters;
    }

    public String[] getHeader() throws IOException {
        return this.reader.getHeader(true);
    }

    public List<String> read() throws IOException {
        return reader.read();
    }

    public void close() throws IOException {
        this.reader.close();
    }

    public List<String> skip(int lines) throws IOException {
        List<String> row = null;
        for ( int i = 0; i < lines; i++ ) {
            row = reader.read();
        }
        return row;
    }

    public CsvPreference getCSVPreferences() {
        try {
            String s = null;
            char quoteChar;
            int delimiterChar;
            String endOfLineSymbols;

            DynObject params = this.getParameters();

            CsvPreference.Builder builder = null;

            CsvPreference defaultPreference = CSVStoreParameters
                    .getPredefinedCSVPreferences(params);
            if ( defaultPreference == null ) {
                defaultPreference = CsvPreference.STANDARD_PREFERENCE;
            }

            endOfLineSymbols = CSVStoreParameters.getRecordSeparator(params);
            if ( StringUtils.isBlank(endOfLineSymbols) ) {
                endOfLineSymbols = defaultPreference.getEndOfLineSymbols();
            }
            s = CSVStoreParameters.getQuoteCharacter(params);
            if ( StringUtils.isBlank(s) ) {
                quoteChar = (char) defaultPreference.getQuoteChar();
            } else {
                quoteChar = s.charAt(0);
            }
            s = CSVStoreParameters.getDelimiter(params);
            if ( StringUtils.isBlank(s) ) {
                delimiterChar = defaultPreference.getDelimiterChar();
            } else {
                delimiterChar = s.charAt(0);
            }

            builder = new CsvPreference.Builder(quoteChar, delimiterChar,
                    endOfLineSymbols);

            s = CSVStoreParameters.getCommentStartMarker(params);
            if ( !StringUtils.isBlank(s) ) {
                CommentStartsWith cs = new CommentStartsWith(s);
                builder.skipComments(cs);
            }

            builder.surroundingSpacesNeedQuotes(CSVStoreParameters
                    .getSurroundingSpacesNeedQuotes(params));
            QuoteMode quoteMode = CSVStoreParameters.getQuoteMode(params);
            if ( quoteMode != null ) {
                builder.useQuoteMode(quoteMode);
            }
            return builder.build();
        } catch (Exception e) {
            logger.warn("Can't make preferences for CSV '" + getFullFileName()
                    + "'.", e);
            return null;
        }
    }
    
    private String getFullFileName() {
        // Usar solo para mostrar mensajes en el logger.
        String s = "(unknow)";
        try {
            s = getParameters().getFile().getAbsolutePath();
        } catch (Exception e2) {
            s = "(unknow)";
        }
        return s;        
    }

    public int getLine() {
        if( this.reader==null ) {
            return 0;
        }
        return this.reader.getLineNumber();
    }

}
