/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.fmap.dal.resource.spi;

import java.util.ArrayList;
import java.util.List;

import org.gvsig.fmap.dal.DALLocator;
import org.gvsig.fmap.dal.exception.InitializeException;
import org.gvsig.fmap.dal.resource.ResourceParameters;
import org.gvsig.fmap.dal.resource.exception.AccessResourceException;
import org.gvsig.fmap.dal.resource.exception.PrepareResourceException;
import org.gvsig.fmap.dal.resource.exception.ResourceException;
import org.gvsig.fmap.dal.resource.exception.ResourceNotifyChangesException;
import org.gvsig.fmap.dal.resource.exception.ResourceNotifyCloseException;
import org.gvsig.fmap.dal.resource.exception.ResourceNotifyDisposeException;
import org.gvsig.fmap.dal.resource.exception.ResourceNotifyOpenException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Resource implementation which is able to show an unique interface to a group
 * of Resources.
 * 
 * @author 2009- <a href="cordinyana@gvsig.org">Csar Ordiana</a> - gvSIG team
 */
public class MultiResource extends AbstractResource {

	public static final String TYPE_NAME = "multi";

	public static final String DESCRIPTION = "Group of multiple resources";

	private static final Logger LOG =
			LoggerFactory.getLogger(MultiResource.class);

	private List resources = new ArrayList();

	// Initially, the first one
	private int defaultResourcePos = 0;

	/**
	 * Creates a new {@link MultiResource} from the given parameters.
	 * 
	 * @param parameters
	 *            to create the resource from
	 * @throws InitializeException
	 *             if there is an error creating the resource
	 */
	public MultiResource(MultiResourceParameters parameters)
			throws InitializeException {
		super(parameters);
	}

	private List getResources() {
		return resources;
	}

	/**
	 * Adds a new {@link ResourceProvider} to the list of Resources managed by
	 * this group.
	 * 
	 * @param parameters
	 *            to create the resource from
	 * @param defaultResource
	 *            if the added resource is to be used as the multi resource
	 *            default one
	 * @throws InitializeException
	 *             if there is an error adding the resource
	 */
	public void addResource(ResourceParameters parameters,
			boolean defaultResource) throws InitializeException {
		ResourceProvider resourceProvider =
				((ResourceManagerProviderServices) DALLocator.getResourceManager()).createResource(parameters);
		resources.add(resourceProvider);
		if (defaultResource) {
			defaultResourcePos = resources.size() - 1;
		}
	}

	/**
	 * Adds a new {@link ResourceProvider} to the list of Resources managed by
	 * this group.
	 * 
	 * @param name
	 *            of the resource to add
	 * @param params
	 *            parameters to create the resource with
	 * @param defaultResource
	 *            if the added resource is to be used as the multi resource
	 *            default one
	 * @throws InitializeException
	 *             if there is an error adding the resource
	 */
	public void addResource(String name, Object[] params,
			boolean defaultResource) throws InitializeException {
		ResourceProvider resourceProvider =
				((ResourceManagerProviderServices) DALLocator.getResourceManager()).createResource(
						name, params);
		resources.add(resourceProvider);
		if (defaultResource) {
			defaultResourcePos = resources.size() - 1;
		}
	}

	public boolean isThis(ResourceParameters parameters)
			throws ResourceException {
		if (parameters instanceof MultiResourceParameters) {
			MultiResourceParameters multiParams =
					(MultiResourceParameters) parameters;
			return multiParams.getName() != null
					&& multiParams.getName().equals(
							getMultiResourceParameters().getName());
		}

		return false;
	}

	public void notifyChanges() throws ResourceNotifyChangesException {
		List resources = getResources();
		for (int i = 0; i < resources.size(); i++) {
			((ResourceProvider) getResources().get(i)).notifyChanges();
		}
	}

	public void notifyClose() throws ResourceNotifyCloseException {
		List resources = getResources();
		for (int i = 0; i < resources.size(); i++) {
			((ResourceProvider) getResources().get(i)).notifyClose();
		}
	}

	public void notifyDispose() throws ResourceNotifyDisposeException {
		List resources = getResources();
		for (int i = 0; i < resources.size(); i++) {
			((ResourceProvider) getResources().get(i)).notifyDispose();
		}
	}

	public void notifyOpen() throws ResourceNotifyOpenException {
		List resources = getResources();
		for (int i = 0; i < resources.size(); i++) {
			((ResourceProvider) getResources().get(i)).notifyOpen();
		}
	}

	public void prepare() throws PrepareResourceException {
		List resources = getResources();
		for (int i = 0; i < resources.size(); i++) {
			((ResourceProvider) getResources().get(i)).prepare();
		}
	}

	public void closeRequest() throws ResourceException {
		List resources = getResources();
		for (int i = 0; i < resources.size(); i++) {
			((ResourceProvider) getResources().get(i)).closeRequest();
		}
	}

	public Object get() throws AccessResourceException {
		ResourceProvider defaultResource = getDefaultResource();
		if (defaultResource == null) {
			return null;
		} else {
			return defaultResource.get();
		}
	}

	public Object getData() {
		ResourceProvider defaultResource = getDefaultResource();
		if (defaultResource == null) {
			return null;
		} else {
			return defaultResource.getData();
		}
	}

	private MultiResourceParameters getMultiResourceParameters() {
		return (MultiResourceParameters) getParameters();
	}

	public String getName() throws AccessResourceException {
		StringBuffer buffer =
				new StringBuffer().append("MultiResource ").append(
						getMultiResourceParameters().getName()).append(":: ");
		List resources = getResources();
		for (int i = 0; i < resources.size(); i++) {
			buffer.append(((ResourceProvider) resources.get(i)).getName());
			if (defaultResourcePos == i) {
				buffer.append(" (default)");
			}
			if (i < resources.size() - 1) {
				buffer.append(", ");
			}
		}
		return buffer.toString();
	}

	public String toString() {
		try {
			return getName();
		} catch (AccessResourceException e) {
			LOG.error("Error getting the resource name", e);
		}
		return super.toString();
	}

	private ResourceProvider getDefaultResource() {
		if (getResources().size() > 0) {
			return (ResourceProvider) getResources().get(defaultResourcePos);
		} else {
			return null;
		}
	}

	public void setData(Object data) {
		ResourceProvider defaultResource = getDefaultResource();
		if (defaultResource != null) {
			defaultResource.setData(data);
		}
	}
}
