/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */

package org.gvsig.fmap.geom.primitive;

import org.gvsig.fmap.geom.DirectPosition;
import org.gvsig.fmap.geom.GeometryException;
import org.gvsig.fmap.geom.aggregate.MultiPoint;

/**
 * <p>
 * This interface is equivalent to the GM_Point specified in 
 * <a href="http://www.iso.org/iso/iso_catalogue/catalogue_tc/catalogue_detail.htm?csnumber=26012">ISO 19107</a>.
 * It is the basic data type for a geometric object consisting 
 * of one and only one point. 
 * </p>
 * @see <a href="http://www.iso.org/iso/iso_catalogue/catalogue_tc/catalogue_detail.htm?csnumber=26012">ISO 19107</a>
 * @author <a href="mailto:jpiera@gvsig.org">Jorge Piera</a>
 */
public interface Point extends Primitive {
	
	/**
	 * Gets the {@link DirectPosition of a point}, that is
	 * composed by a set of ordinates
	 * @return
	 * The direct position
	 */
	public DirectPosition getDirectPosition();
	
	/**
	 * Sets a ordinate in a concrete dimension
	 * @param dimension
	 * The dimension to set
	 * @param value
	 * The value to set
	 */
	void setCoordinateAt(int dimension, double value);

    /**
     * Sets the point coordinates
     * 
     * @param value
     *            The coordinates to set
     * @deprecated Use {@link #setCoordinateAt(int, double))} instead
     */
	void setCoordinates(double[] values);
	
	/**
	 * Sets the X coordinate
	 * @param x
	 * The value to set
	 */
	void setX(double x);
	
	/**
	 * Sets the Y coordinate
	 * @param y
	 * The value to set
	 */
	void setY(double y);
	
	/**
	 * Gets the coordinate in a concrete dimension
	 * @param dimension
	 * The ordinate dimension
	 * @return
	 * The value of the ordinate
	 */
	double getCoordinateAt(int dimension);

    /**
     * Returns an array of coordinates. Don't use the provided array to modify
     * the point coordinates, but use the {@link #setCoordinateAt(int, double))}
     * , {@link #setX(double))} or {@link #setY(double))} methods instead.
     * 
     * @return The point coordinates
     * @deprecated Use {@link #getCoordinateAt(int)} instead
     */
	double[] getCoordinates();
	
	/**
	 * Returns the X coordinate 
	 * @return
	 * The X coordinate
	 */
	double getX();
	
	/**
	 * Returns the Y coordinate 
	 * @return
	 * The Y coordinate
	 */
	double getY();
	
        public MultiPoint toPoints() throws GeometryException ;
}
