/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.fmap.mapcontrol.tools.Behavior;

import java.awt.Color;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;

import org.gvsig.fmap.geom.Geometry.SUBTYPES;
import org.gvsig.fmap.geom.GeometryLocator;
import org.gvsig.fmap.geom.GeometryManager;
import org.gvsig.fmap.geom.exception.CreateEnvelopeException;
import org.gvsig.fmap.geom.primitive.Envelope;
import org.gvsig.fmap.mapcontext.ViewPort;
import org.gvsig.fmap.mapcontrol.MapControlDrawer;
import org.gvsig.fmap.mapcontrol.tools.BehaviorException;
import org.gvsig.fmap.mapcontrol.tools.Events.EnvelopeEvent;
import org.gvsig.fmap.mapcontrol.tools.Listeners.RectangleListener;
import org.gvsig.fmap.mapcontrol.tools.Listeners.ToolListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;



/**
 * <p>Behavior that permits user to select a rectangular area on the associated <code>MapControl</code> using
 *  a {@link RectangleListener RectangleListener}.</p>
 *
 * @author Vicente Caballero Navarro
 * @author Pablo Piqueras Bartolom
 */
public class RectangleBehavior extends Behavior {
	private static final GeometryManager geomManager = GeometryLocator.getGeometryManager();
	private static final Logger logger = LoggerFactory.getLogger(RectangleBehavior.class);

	/**
	 * First point of the rectangle area, that represents a corner.
	 */
	private Point2D m_FirstPoint;

	/**
	 * Second point of the rectangle area, that represents the opposite corner of the first,
	 *  along the diagonal.
	 */
	private Point2D m_LastPoint;

	/**
	 * Tool listener used to work with the <code>MapControl</code> object.
	 *
	 * @see #getListener()
	 * @see #setListener(ToolListener)
	 */
	private RectangleListener listener;

	/**
	 * <p>Creates a new behavior for selecting rectangle areas.</p>
	 *
	 * @param zili listener used to permit this object to work with the associated <code>MapControl</code>
	 */
	public RectangleBehavior(RectangleListener zili) {
		listener = zili;
	}

	/*
	 * (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.tools.Behavior.Behavior#paintComponent(java.awt.Graphics)
	 */
	public void paintComponent(MapControlDrawer mapControlDrawer) {
		mapControlDrawer.setColor(Color.black);

		// Borramos el anterior
		Rectangle r = new Rectangle();

		// Dibujamos el actual
		if ((m_FirstPoint != null) && (m_LastPoint != null)) {
			r.setFrameFromDiagonal(m_FirstPoint, m_LastPoint);
			mapControlDrawer.drawRect(r.x, r.y, r.width, r.height);
		}
	}

	/*
	 * (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.tools.Behavior.Behavior#mousePressed(java.awt.event.MouseEvent)
	 */
	public void mousePressed(MouseEvent e) {
            if( !isMyButton(e) ) {
                return;
            }
		if (e.getButton() == MouseEvent.BUTTON1) {
			m_FirstPoint = e.getPoint();
			getMapControl().repaint();
		}
		if (listener.cancelDrawing()) {
			getMapControl().cancelDrawing();
			getMapControl().repaint();
		}
	}

	/*
	 * (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.tools.Behavior.Behavior#mouseReleased(java.awt.event.MouseEvent)
	 */
	public void mouseReleased(MouseEvent e) throws BehaviorException {
            if( !isMyButton(e) ) {
                return;
            }
	    if (m_FirstPoint == null) return;
		Point2D p1;
		Point2D p2;
		Point pScreen = e.getPoint();

		ViewPort vp = getMapControl().getMapContext().getViewPort();

		p1 = vp.toMapPoint(m_FirstPoint);
		p2 = vp.toMapPoint(pScreen);

		if (e.getButton() == MouseEvent.BUTTON1) {
			//	Fijamos el nuevo extent

			double x=0;
			double y=0;
			double x1=0;
			double y1=0;
			if (p1.getX()<p2.getX()){
				x=p1.getX();
				x1=p2.getX();
			}else{
				x=p2.getX();
				x1=p1.getX();
			}
			if (p1.getY()<p2.getY()){
				y=p1.getY();
				y1=p2.getY();
			}else{
				y=p2.getY();
				y1=p1.getY();
			}
			Envelope r;
			try {
				r = geomManager.createEnvelope(x,y,x1,y1, SUBTYPES.GEOM2D);
				Rectangle2D rectPixel = new Rectangle();
				rectPixel.setFrameFromDiagonal(m_FirstPoint, pScreen);

				EnvelopeEvent event = new EnvelopeEvent(r, e, rectPixel);
				listener.rectangle(event);
			} catch (CreateEnvelopeException e1) {
				logger.error("Error creating the envelope", e);
			}//Rectangle2D.Double();

			//			r.setFrameFromDiagonal(p1, p2);


		}

		m_FirstPoint = null;
		m_LastPoint = null;
	}

	/*
	 * (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.tools.Behavior.Behavior#mouseDragged(java.awt.event.MouseEvent)
	 */
	public void mouseDragged(MouseEvent e) {
            if( !isMyButton(e) ) {
                return;
            }
		m_LastPoint = e.getPoint();
		getMapControl().repaint();
	}

	/**
	 * <p>Sets a tool listener to work with the <code>MapControl</code> using this behavior.</p>
	 *
	 * @param listener a <code>RectangleListener</code> object for this behavior
	 */
	public void setListener(ToolListener listener) {
		this.listener = (RectangleListener) listener;
	}

	/*
	 * (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.tools.Behavior.Behavior#getListener()
	 */
	public ToolListener getListener() {
		return listener;
	}
}
