/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.utils.swing.wizard;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.util.ArrayList;
import java.util.Iterator;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JPanel;

import org.gvsig.tools.swing.api.ToolsSwingLocator;

/**
 * Clase wizard con la gestin de los botones y los paneles aadidos al mismo.
 * Al avanzar o retrasar un paso el asistente se desactivan todos los botones
 * menos el de cancelar, que est activo siempre por defecto
 *
 * @author Fernando Gonzlez Corts
 */
public class Wizard extends JPanel implements WizardControl {
	private JPanel jPanel = null;
	private JButton btnBack = null;
	private JButton btnNext = null;
	private JButton btnFinish = null;
	private JButton btnCancel = null;
	private ArrayList steps = new ArrayList();
	private int currentStep = 0;
	private JPanel pnlSteps = null;
	private ArrayList listeners = new ArrayList();

	/**
	 * This is the default constructor
	 */
	public Wizard(String backText, String nextText, String finishText, String cancelText) {
		super();
		initialize(backText, nextText, finishText, cancelText);
	}

	/**
	 * Aade un lstener de eventos del wizard
	 *
	 * @param listener
	 */
	public void addWizardListener(WizardListener listener) {
		listeners.add(listener);
	}

	/**
	 * Elimina un listener de eventos del wizard
	 *
	 * @param listener
	 */
	public void removeWizardListener(WizardListener listener) {
		listeners.remove(listener);
	}

	/**
	 * Invoca en los listeners el evento cancel
	 */
	private void callCancelListener() {
		for (Iterator iter = listeners.iterator(); iter.hasNext();) {
			WizardListener element = (WizardListener) iter.next();
			element.cancel(new WizardEvent(this, currentStep));
		}
	}

	/**
	 * Invoca en los listeners el evento finish
	 */
	private void callFinishListener() {
		for (Iterator iter = listeners.iterator(); iter.hasNext();) {
			WizardListener element = (WizardListener) iter.next();
			element.finished(new WizardEvent(this, currentStep));
		}
	}

	/**
	 * Invoca en los listeners el evento next
	 */
	private void callNextListener() {
		for (Iterator iter = listeners.iterator(); iter.hasNext();) {
			WizardListener element = (WizardListener) iter.next();
			element.next(new WizardEvent(this, currentStep));
		}
	}

	/**
	 * Invoca en los listeners el evento back
	 */
	private void callBackListener() {
		for (Iterator iter = listeners.iterator(); iter.hasNext();) {
			WizardListener element = (WizardListener) iter.next();
			element.back(new WizardEvent(this, currentStep));
		}
	}

	/**
	 * This method initializes this
	 */
	private void initialize(String backText, String nextText, String finishText, String cancelText) {
		this.setLayout(new BorderLayout());
		this.setSize(300, 200);
		this.add(getJPanel(backText, nextText, finishText, cancelText), java.awt.BorderLayout.SOUTH);
		this.add(getPnlSteps(), java.awt.BorderLayout.CENTER);
	}

	/**
	 * Aade un paso al asistente. Inicializa el paso
	 *
	 * @param s Paso a aadir
	 *
	 * @throws RuntimeException DOCUMENT ME!
	 */
	public void addStep(Step s) {
		if (!(s instanceof JComponent)) {
			throw new RuntimeException(
				"Step must be a visual component (descend from JComponent)");
		}

		getPnlSteps().add((JComponent) s, BorderLayout.CENTER);
		steps.add(s);
		disableButtons();
		s.init(this);
	}

	/**
	 * Habilita o deshabilita los botones en funcin del paso en el que se
	 * encuentre el asistente
	 */
	private void disableButtons() {
		btnNext.setEnabled(false);
		btnBack.setEnabled(false);
		btnFinish.setEnabled(false);
	}

	/**
	 * Activa el paso al siguiente paso del asistente
	 */
	public void enableNext(boolean enabled) {
		if (currentStep == (steps.size() - 1)) {
			btnFinish.setEnabled(enabled);
		} else {
			btnNext.setEnabled(enabled);
		}
	}

	/**
	 * Activa el paso al paso anterior del asistente
	 */
	public void enableBack(boolean enabled) {
		if (currentStep != 0) {
			btnBack.setEnabled(enabled);
		}
	}

	/**
	 * This method initializes jPanel
	 *
	 * @return javax.swing.JPanel
	 */
	private JPanel getJPanel(String backText, String nextText, String finishText, String cancelText) {
		if (jPanel == null) {
			jPanel = new JPanel();
			jPanel.add(getBtnBack(backText), null);
			jPanel.add(getBtnNext(nextText), null);
			jPanel.add(getBtnFinish(finishText), null);
			jPanel.add(getBtnCancel(cancelText), null);
		}

		return jPanel;
	}

	private JButton newJButton(String text){
        return ToolsSwingLocator.getUsabilitySwingManager().createJButton(text);
	}
	
	/**
	 * Obtiene una referencia al botn de dar un paso atrs
	 * @param text
	 *
	 * @return javax.swing.JButton
	 */
	public JButton getBtnBack(String text) {
		if (btnBack == null) {
			btnBack = newJButton(text);
			btnBack.setText(text);
			btnBack.setMargin(new java.awt.Insets(2, 2, 2, 2));
			btnBack.setEnabled(false);
			btnBack.addActionListener(new java.awt.event.ActionListener() {
					public void actionPerformed(java.awt.event.ActionEvent e) {
						backStep();
					}
				});
		}

		return btnBack;
	}

	/**
	 * Obtiene una referencia al botn de dar un paso adelante
	 * @param text
	 *
	 * @return javax.swing.JButton
	 */
	public JButton getBtnNext(String text) {
		if (btnNext == null) {
			btnNext = newJButton(text);
			btnNext.setText(text);
			btnNext.setMargin(new java.awt.Insets(2, 2, 2, 2));
			btnNext.setEnabled(false);
			btnNext.addActionListener(new java.awt.event.ActionListener() {
					public void actionPerformed(java.awt.event.ActionEvent e) {
						nextStep();
					}
				});
		}

		return btnNext;
	}

	/**
	 * Obtiene una referencia al botn de finalizar
	 * @param finishText
	 *
	 * @return javax.swing.JButton
	 */
	public JButton getBtnFinish(String text) {
		if (btnFinish == null) {
			btnFinish = newJButton(text);
			btnFinish.setMargin(new java.awt.Insets(2, 2, 2, 2));
			btnFinish.setText(text);
			btnFinish.setEnabled(false);
			btnFinish.addActionListener(new java.awt.event.ActionListener() {
					public void actionPerformed(java.awt.event.ActionEvent e) {
						callFinishListener();
					}
				});
		}

		return btnFinish;
	}

	/**
	 * Obtiene una referencia al botn de cancelar
	 * @param cancelText
	 *
	 * @return javax.swing.JButton
	 */
	public JButton getBtnCancel(String text) {
		if (btnCancel == null) {
			btnCancel = newJButton(text);
			btnCancel.setMargin(new java.awt.Insets(2, 2, 2, 2));
			btnCancel.setText(text);
			btnCancel.addActionListener(new java.awt.event.ActionListener() {
					public void actionPerformed(java.awt.event.ActionEvent e) {
						callCancelListener();
					}
				});
		}

		return btnCancel;
	}

	/**
	 * This method initializes pnlSteps
	 *
	 * @return javax.swing.JPanel
	 */
	private JPanel getPnlSteps() {
		if (pnlSteps == null) {
			pnlSteps = new JPanel();
			pnlSteps.setLayout(new CardLayout());
		}

		return pnlSteps;
	}

	/**
	 * Muestra el panel del siguiente paso del asistente
	 */
	public void nextStep() {
		currentStep++;
		((CardLayout) getPnlSteps().getLayout()).next(getPnlSteps());
		disableButtons();
		callNextListener();
	}

	/**
	 * Muestra el panel del paso anterior del asistente
	 */
	public void backStep() {
		currentStep--;
		((CardLayout) getPnlSteps().getLayout()).previous(getPnlSteps());
		disableButtons();
		callBackListener();
	}

	/**
	 * Se cancela el asistente. Esta operacin no tiene ningn efecto, salvo
	 * que se disparar el evento de cancelado. El resultado de esto depender
	 * de las implementaciones que haya escuchando el evento. Generalmente
	 * deber haber un objeto que al escuchar este evento cerrar el
	 * asistente.
	 */
	public void cancel() {
		callCancelListener();
	}

	/**
	 * Se finaliza el asistente. Esta operacin no tiene ningn efecto, salvo
	 * que se disparar el evento de finalizacin. El resultado de esto
	 * depender de las implementaciones que haya escuchando el evento.
	 * Generalmente deber haber un objeto que al escuchar este evento cerrar
	 * el asistente.
	 */
	public void finish() {
		callFinishListener();
	}

	/**
	 * Obtiene un array con los pasos del asistente
	 *
	 * @return array de pasos
	 */
	public Step[] getSteps() {
		return (Step[]) steps.toArray(new Step[0]);
	}

	/**
	 * Obtiene el paso actual del asistente
	 *
	 * @return Paso actual del asistente
	 */
	public Step getCurrentStep() {
		return getSteps()[currentStep];
	}
}
