/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
/*
 * AUTHORS (In addition to CIT):
 * 2008 {DiSiD Technologies}  {New extension for installation and update of text translations}
 */
package org.gvsig.i18n.impl;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;
import org.gvsig.andami.LocaleManager;
import org.gvsig.andami.PluginsLocator;
import org.gvsig.i18n.I18nException;
import org.gvsig.i18n.I18nManager;
import org.gvsig.i18n.Messages;
import org.gvsig.utils.StringUtilities;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Implementation of the I18nManager interface.
 *
 * @author <a href="mailto:dcervera@disid.com">David Cervera</a>
 * @author <a href="mailto:cordin@disid.com">Csar Ordiana</a>
 */
public class I18nManagerImpl implements I18nManager {

    private static final Logger logger = LoggerFactory.getLogger(I18nManagerImpl.class);
    
    private static final String LOCALES_FILE_NAME = "locales.csv";

    private static final String CSV_SEPARATOR = ",";

    private static final I18nManager DEFAULT = new I18nManagerImpl();

    /**
     * The list of default reference locales. The last one will be used to get
     * all the keys when translating to a new locale.
     */
    private Locale[] referenceLocales = new Locale[]{ENGLISH, SPANISH};

    LocaleManager localeManager = null;

    private LocaleManager getLocaleManager() {
        if ( localeManager == null ) {
            localeManager = PluginsLocator.getLocaleManager();
        }
        return localeManager;
    }

    /**
     * Returns the unique instance of the I18nManager.
     *
     * @return the unique instance
     */
    public static I18nManager getInstance() {
        return DEFAULT;
    }

    public static String capitalize(String text) {
        // Convert the first letter to uppercase
        String capitalLetter
                = new String(new char[]{Character.toUpperCase(text.charAt(0))});
        return capitalLetter.concat(text.substring(1));
    }

    private Set<Locale> getRegisteredLocales() {
        return getLocaleManager().getInstalledLocales();
    }

    public Locale getCurrentLocale() {
        return getLocaleManager().getCurrentLocale();
    }

    public Locale[] getInstalledLocales() {
        Set<Locale> locales = getRegisteredLocales();
        return (Locale[]) locales.toArray(new Locale[locales.size()]);
    }

    private File getResourcesFolder() {
        return getLocaleManager().getResourcesFolder();
    }

    public void setCurrentLocale(Locale locale) {
        getLocaleManager().setCurrentLocale(locale);
    }

    public Locale getDefaultSystemLocale() {
        return getLocaleManager().getDefaultSystemLocale();
    }

    public void installLocale(Locale locale) {
        getLocaleManager().uninstallLocale(locale);
    }

    public void uninstallLocale(Locale locale) throws I18nException {
        if ( getCurrentLocale().equals(locale) || isReferenceLocale(locale) ) {
            throw new UninstallLocaleException(locale);
        }
        if ( getLocaleManager().uninstallLocale(locale) ) {
            // Remove the resource bundle file
            File bundleFile
                    = new File(getResourcesFolder(), getResourceFileName(locale));

            if ( bundleFile.exists() ) {
                bundleFile.delete();
            }
        }
    }

    public String getDisplayName(Locale locale) {
        return getDisplayName(locale, locale);
    }

    public String getDisplayName(Locale locale, Locale displayLocale) {
        StringBuffer name
                = new StringBuffer(getLanguageDisplayName(locale, displayLocale));

        if ( !isEmpty(locale.getCountry()) ) {
            name.append(" - ");
            name.append(locale.getDisplayCountry(displayLocale));
        }

        if ( !isEmpty(locale.getVariant()) ) {
            name.append(" - ");
            name.append(locale.getDisplayVariant(displayLocale));
        }

        name.append(" (").append(locale.toString()).append(")");

        return name.toString();
    }

    private boolean isEmpty(String text) {
        return text == null || text.trim().length() == 0;
    }

    public String getLanguageDisplayName(Locale locale) {
        return this.getLocaleManager().getLanguageDisplayName(locale);
    }

    public String getLanguageDisplayName(Locale locale, Locale displayLocale) {
        return this.getLocaleManager().getLanguageDisplayName(locale);
        /*
        String displayName;

	// Correction for the Basque language display name,
        // show it in Basque, not in Spanish
        if ( "eu".equals(locale.getLanguage())
                && "vascuence".equals(locale.getDisplayLanguage()) ) {
            displayName = "Euskera";
        } // Patch for Valencian/Catalan
        else if ( "ca".equals(locale.getLanguage()) ) {
	    // displayName = Messages.getText("__valenciano");
            // if ("__valenciano".equals(displayName)) {
            // displayName = Messages.getText("__catalan");
            // }
            displayName = "Valenci";
        } else {
            displayName = locale.getDisplayLanguage(displayLocale);
        }

        return capitalize(displayName);
        */
    }

    public Locale[] installLocales(File importFile) throws I18nException {
        List importLocales = new ArrayList();

        try {
            ZipFile zipFile = new ZipFile(importFile);

            Map locales = getZipFileNonReferenceLocales(zipFile);

            if ( locales == null || locales.size() == 0 ) {
                return null;
            }

            for ( Iterator iterator = locales.entrySet().iterator(); iterator.hasNext(); ) {
                Entry entry = (Entry) iterator.next();

                String fileName = (String) entry.getKey();
                Locale locale = (Locale) entry.getValue();
                importLocales.add(locale);

                getLocaleManager().installLocale(locale);

                // Replace the old bundle with the new one
                ZipEntry zipEntry = zipFile.getEntry(fileName);
                InputStream is = zipFile.getInputStream(zipEntry);
                BufferedReader reader
                        = new BufferedReader(new InputStreamReader(is));

                File bundleFile = getResourceFile(locale);
                FileWriter fileWriter = new FileWriter(bundleFile);
                BufferedWriter writer = new BufferedWriter(fileWriter);

                String line;
                while ( (line = reader.readLine()) != null ) {
                    writer.write(line);
                    writer.write('\n');
                }
                writer.flush();
                writer.close();
                fileWriter.close();
                reader.close();
                is.close();
            }

        } catch (Exception ex) {
            throw new InstallLocalesException(importFile, ex);
        }

        return (Locale[]) importLocales.toArray(new Locale[importLocales.size()]);
    }

    public void exportLocaleForUpdate(Locale locale, Locale referenceLocale,
            File exportFile) throws I18nException {

        exportLocalesForUpdate(new Locale[]{locale}, referenceLocale,
                exportFile);
    }

    public void exportLocalesForUpdate(Locale[] locales,
            Locale referenceLocale, File exportFile) throws I18nException {

        exportLocale(locales, new Locale[]{referenceLocale}, exportFile,
                true);
    }

    public void exportLocaleForTranslation(Locale locale,
            Locale referenceLocale, File exportFile) throws I18nException {

        exportLocaleForTranslation(locale, new Locale[]{referenceLocale},
                exportFile);
    }

    public void exportLocaleForTranslation(Locale locale,
            Locale[] referenceLocales, File exportFile) throws I18nException {

        exportLocale(new Locale[]{locale}, referenceLocales, exportFile,
                false);
    }

    public Locale[] getReferenceLocales() {
        return referenceLocales;
    }

    public void setReferenceLocales(Locale[] referenceLocales) {
        this.referenceLocales = referenceLocales;
    }

    public void setDefaultLocales(Locale[] defaultLocales) {
        logger.warn("This operation (setDefaultLocales) is not supported.");
    }

    private void exportLocale(Locale[] locales, Locale[] referenceLocales,
            File exportFile, boolean update) throws I18nException {

        Locale[] refArray
                = getReferenceLocalesToExport(locales, referenceLocales);

        Set allReferenceKeys = new HashSet();
        Map referenceTexts = new HashMap(refArray.length);
        for ( int i = 0; i < refArray.length; i++ ) {
            Map texts = getAllTexts(refArray[i]);
            referenceTexts.put(refArray[i], texts);
            allReferenceKeys.addAll(texts.keySet());
        }

        try {
            FileOutputStream fos = new FileOutputStream(exportFile);
            ZipOutputStream zipos = new ZipOutputStream(fos);

            // Create the index file
            writeZipFileLocales(zipos, locales, refArray);

            PrintStream ps = new PrintStream(zipos);
            Map texts = null;

            if ( update ) {
                // First, export the locales to update
                if ( locales != null ) {
                    for ( int i = 0; i < locales.length; i++ ) {
                        texts = getAllTexts(locales[i]);
                        addPendingKeys(texts, allReferenceKeys);
                        putResourceInZip(zipos, ps, texts,
                                getResourceFileName(locales[i]));
                    }
                }
            } else { // translate
                // First, export the locales to translate, taking the keys from
                // the reference locales, but without values
                if ( locales != null ) {
                    for ( int i = 0; i < locales.length; i++ ) {
                        putResourceInZip(zipos, ps, allReferenceKeys,
                                getResourceFileName(locales[i]));
                    }
                }
            }

            // Next, export the reference locales
            if ( refArray != null ) {
                for ( int i = 0; i < refArray.length; i++ ) {
                    texts = getAllTexts(refArray[i]);
                    putResourceInZip(zipos, ps, texts,
                            getResourceFileName(refArray[i]));
                }
            }

            ps.flush();
            ps.close();
            zipos.close();
            fos.close();
        } catch (IOException ex) {
            throw new ExportLocaleException(locales, ex);
        }
    }

    /**
     * Adds the keys of the set to the map that aren't still contained in it
     * with a related empty String.
     *
     * @param texts
     * the map to complete with the keys
     * @param allReferenceKeys
     * the complete key set
     */
    private void addPendingKeys(Map texts, Set allReferenceKeys) {
        for ( Iterator iterator = allReferenceKeys.iterator(); iterator.hasNext(); ) {
            Object key = (Object) iterator.next();
            if ( !texts.containsKey(key) ) {
                texts.put(key, "");
            }
        }
    }

    /**
     * Returns the list of reference locales to export, as the union of the
     * default reference locales list and the one selected as reference. The
     * locales to translate or update are extracted from the list.
     */
    private Locale[] getReferenceLocalesToExport(Locale[] locales,
            Locale[] referenceLocalesSelected) {
		// The reference locales to export are the default ones plus the
        // selected by the user.
        Set exportRefLocales = new HashSet(referenceLocales.length);
        for ( int i = 0; i < referenceLocales.length; i++ ) {
            exportRefLocales.add(referenceLocales[i]);
        }
        if ( referenceLocalesSelected != null ) {
            for ( int i = 0; i < referenceLocalesSelected.length; i++ ) {
                exportRefLocales.add(referenceLocalesSelected[i]);
            }
        }
        if ( locales != null ) {
            for ( int i = 0; i < locales.length; i++ ) {
                exportRefLocales.remove(locales[i]);
            }
        }
        Locale[] refArray
                = (Locale[]) exportRefLocales.toArray(new Locale[exportRefLocales.size()]);
        return refArray;
    }

    /**
     * Returns all the localized texts and its keys for a locale.
     */
    private Map getAllTexts(Locale locale) {
        return Messages.getAllTexts(locale);
    }

    private Map getZipFileNonReferenceLocales(ZipFile zipFile)
            throws I18nException {
        ZipEntry zipEntry = zipFile.getEntry(LOCALES_FILE_NAME);

        if ( zipEntry == null ) {
            return null;
        }

        Map locales;
        try {
            InputStream is = zipFile.getInputStream(zipEntry);
            BufferedReader reader
                    = new BufferedReader(new InputStreamReader(is));

            locales = new HashMap(2);
            String line;
            while ( (line = reader.readLine()) != null ) {
				// The excepted format is:
                // FILENAME,LANGUAGE,COUNTRY,VARIANT,IS_REFERENCE
                StringTokenizer st
                        = new StringTokenizer(line, CSV_SEPARATOR, true);
                // First: locale file name (required)
                String fileName = st.nextToken();
                if ( CSV_SEPARATOR.equals(fileName) ) {
                    throw new LocaleFileNameRequiredException(line);
                } else {
                    // Read the next separator
                    st.nextToken();
                }
                // Second: the locale language (required)
                String language = st.nextToken();
                if ( CSV_SEPARATOR.equals(language) ) {
                    throw new LocaleLanguageRequiredException(line);
                } else {
                    // Read the next separator
                    st.nextToken();
                }
                // Third: the country
                String country = st.nextToken();
                if ( CSV_SEPARATOR.equals(country) ) {
                    country = null;
                } else {
                    // Read the next separator
                    st.nextToken();
                }
                // Fourth: the variant
                String variant = st.nextToken();
                if ( CSV_SEPARATOR.equals(variant) ) {
                    variant = null;
                } else {
                    // Read the next separator
                    st.nextToken();
                }
                // Fifth: is a reference locale?
                String refStr = st.nextToken();
                if ( CSV_SEPARATOR.equals(refStr) ) {
                    refStr = null;
                }

                // Only add non reference locales
                if ( refStr != null && !"true".equals(refStr.toLowerCase()) ) {
                    // Variant only accepted if country defined
                    if ( country == null ) {
                        variant = null;
                    }
                    country = country == null ? "" : country;
                    variant = variant == null ? "" : variant;
                    Locale locale = new Locale(language, country, variant);

                    locales.put(fileName, locale);
                }
            }

            reader.close();
            is.close();
        } catch (IOException ex) {
            throw new ReadCSVLocalesFileException(ex);
        }

        return locales;
    }

    private void writeZipFileLocales(ZipOutputStream zos, Locale[] locales,
            Locale[] referenceLocales) throws IOException {
        ZipEntry zipEntry = new ZipEntry(LOCALES_FILE_NAME);

        zos.putNextEntry(zipEntry);
        BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(zos));

        if ( locales != null ) {
            for ( int i = 0; i < locales.length; i++ ) {
                writeLocaleEntry(locales[i], writer, false);
            }
        }
        if ( referenceLocales != null ) {
            for ( int i = 0; i < referenceLocales.length; i++ ) {
                writeLocaleEntry(referenceLocales[i], writer, true);
            }
        }

        writer.flush();
        zos.closeEntry();
    }

    /**
     * Writes the locale entry into a writer.
     *
     * @param locale
     * the locale to create the entry for
     * @param writer
     * to write to
     * @param reference
     * is it is a reference locale or not
     * @throws IOException
     * if there is an error creating the locale entry
     */
    private void writeLocaleEntry(Locale locale, BufferedWriter writer,
            boolean reference) throws IOException {
        String language = locale.getLanguage();
        String country = locale.getCountry();
        country = country == null ? "" : country;
        String variant = locale.getVariant();
        variant = variant == null ? "" : variant;

        writer.write(getResourceFileName(locale));
        writer.write(',');
        writer.write(language);
        writer.write(',');
        writer.write(country);
        writer.write(',');
        writer.write(variant);
        writer.write(',');
        writer.write(Boolean.toString(reference));
        writer.write('\n');
    }

    /**
     * Returns if a locale is one of the default reference ones.
     */
    private boolean isReferenceLocale(Locale locale) {
        for ( int i = 0; i < referenceLocales.length; i++ ) {
            if ( referenceLocales[i].equals(locale) ) {
                return true;
            }
        }
        return false;
    }

    /**
     * Puts a new resource file into a Jar file.
     */
    private void putResourceInZip(ZipOutputStream zipos, PrintStream ps,
            Map texts, String resourceFileName) throws IOException {
        // Add ZIP entry for the resource bundle file
        zipos.putNextEntry(new ZipEntry(resourceFileName));

        for ( Iterator iterator = texts.entrySet().iterator(); iterator.hasNext(); ) {
            Entry entry = (Entry) iterator.next();
            String keyEncoded = escape((String) entry.getKey(), true);
            ps.print(keyEncoded);
            ps.print("=");
            String valueEncoded = escape((String) entry.getValue(), false);
            ps.println(valueEncoded);
        }

        ps.flush();

        // Close the ZIP entry, the file is complete
        zipos.closeEntry();
    }

    /**
     * Puts a new resource file into a Jar file.
     */
    private void putResourceInZip(ZipOutputStream zipos, PrintStream ps,
            Set keys, String resourceFileName) throws IOException {
        // Add ZIP entry for the resource bundle file
        zipos.putNextEntry(new ZipEntry(resourceFileName));

        for ( Iterator iterator = keys.iterator(); iterator.hasNext(); ) {
            String value = (String) iterator.next();
            String keyEncoded = escape(value, true);
            ps.print(keyEncoded);
            ps.print("=");
            ps.println();
        }

        ps.flush();

        // Close the ZIP entry, the file is complete
        zipos.closeEntry();
    }

    /**
     * Returns the file which contains the translations for a locale.
     */
    private File getResourceFile(Locale locale) {
        return new File(getResourcesFolder(), getResourceFileName(locale));
    }

    /**
     * Returns the name of the file which contains the translations for a
     * locale.
     */
    private String getResourceFileName(Locale locale) {
        StringBuffer fileName = new StringBuffer("text");

        // Spanish without country is the default locale
        if ( !(isEmpty(locale.getCountry()) && "es".equals(locale.getLanguage())) ) {
            fileName.append('_').append(locale.getLanguage());
        }

        // Add the locale country
        if ( !isEmpty(locale.getCountry()) ) {
            fileName.append('_').append(locale.getCountry());
        }

        // Add the locale variant
        if ( !isEmpty(locale.getVariant()) ) {
            fileName.append('_').append(locale.getVariant());
        }

        fileName.append(".properties");
        return fileName.toString();
    }

    private String escape(String value, boolean replaceBlanks) {
        // First replace non printable characters
        if ( replaceBlanks ) {
            value = StringUtilities.replace(value, " ", "\\ ");
        }
        value = StringUtilities.replace(value, ":", "\\:");
        value = StringUtilities.replace(value, "\n", "\\n");
        value = StringUtilities.replace(value, "\t", "\\t");
        value = StringUtilities.replace(value, "\b", "\\b");
        value = StringUtilities.replace(value, "\f", "\\f");
        value = StringUtilities.replace(value, "\r", "\\r");
		// value = StringUtilities.replace(value, "\\", "\\\\");
        // value = StringUtilities.replace(value, "\'", "\\\'");
        // value = StringUtilities.replace(value, "\"", "\\\"");

        // Next, encode in raw-unicode-escape
        return toRawUnicodeEncoded(value);
    }

    private String toRawUnicodeEncoded(String value) {
        StringBuffer sb = new StringBuffer();
        for ( int i = 0; i < value.length(); i++ ) {
            char c = value.charAt(i);
            if ( c <= 0x80 ) {
                sb.append(c);
            } else {
                sb.append("\\u");
                String hexValue = Integer.toHexString((int) c);
                // Append 0 if the hex value has less than 4 digits
                for ( int j = hexValue.length(); j < 4; j++ ) {
                    sb.append('0');
                }
                sb.append(hexValue);
            }
        }
        return sb.toString();
    }
}
